# This file is part of the PostgreSQL k8s Charm for Juju.
# Copyright 2020 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
from pathlib import Path
from typing import Iterable

from charmhelpers.core import host
import kubernetes
from kubernetes.client.rest import ApiException as K8sApiException
import ops.framework
import ops.model

from pgconnstr import ConnectionString
from leadership import RichLeaderData
import pg


PG_MAJOR = 12


log = logging.getLogger(__name__)


def get_namespace():
    return open("/var/run/secrets/kubernetes.io/serviceaccount/namespace", "r").read().strip()


class ClientRelations(ops.framework.Object):
    def __init__(self, charm: ops.charm.CharmBase, key: str):
        super().__init__(charm, key)
        self.charm = charm
        self.passwords = RichLeaderData(self, "passwords")

        self.unit = self.model.unit
        self.app = self.model.app

        self.framework.observe(charm.on["db"].relation_changed, self.on_db_relation_changed)
        self.framework.observe(charm.on["db-admin"].relation_changed, self.on_db_admin_relation_changed)

    _authed = False

    def k8s_auth(self):
        if self._authed:
            return

        _fix_lp_1892255()

        kubernetes.config.load_incluster_config()
        self._authed = True

    @property
    def master_service_name(self) -> str:
        return f"{self.app.name}-master"

    @property
    def standbys_service_name(self) -> str:
        return f"{self.app.name}-standbys"

    @property
    def master_service_ip(self) -> str:
        svc = self.get_k8s_service(self.master_service_name)
        if svc is None:
            return None
        return svc.spec.cluster_ip

    @property
    def standbys_service_ip(self) -> str:
        svc = self.get_k8s_service(self.standbys_service_name)
        if svc is None:
            return None
        return svc.spec.cluster_ip

    def get_k8s_service(self, name: str):
        self.k8s_auth()
        api = _core_v1_api()
        try:
            return api.read_namespaced_service(name, get_namespace())
        except K8sApiException as e:
            if e.status == 404:
                return None
            raise

    def on_db_admin_relation_changed(self, event: ops.charm.RelationEvent):
        self.on_db_relation_changed(event, admin=True)

    def on_db_relation_changed(self, event: ops.charm.RelationEvent, admin: bool = False):
        # Database username is the remote Application name.
        username = event.app.name

        password = self.db_password(username)
        master_ip = self.master_service_ip
        standbys_ip = self.standbys_service_ip
        if not password or master_ip is None or standbys_ip is None:
            # Leader will set the password and create the service in
            # the first hook. But maybe invalid config caused
            # pod-set-spec to not be run yet, or similar.
            log.info("Waiting for leader")
            event.defer()  # Wait for leader to choose the password.
            return

        req = _ClientRequests(event)

        log.info(f"Client requested {req.database=} {req.roles=} {req.extensions=}")

        # Fall back to a database named after the remote Application.
        # This is problematic for cross-model relations, where the
        # remote Application name may have been anonymized and we
        # get a unique database name every fresh deployment.
        dbname = req.database or event.app.name

        is_leader = self.unit.is_leader()
        if is_leader:
            con = pg.connect(
                ConnectionString(
                    host=master_ip, dbname="postgres", user="postgres", password=self.charm.get_admin_password()
                )
            )
            pg.ensure_user(con, username, password, superuser=admin)
            pg.ensure_db(con, dbname, username)
            pg.ensure_roles(con, req.roles)
            pg.ensure_extensions(con, req.extensions)

        # Publish allowed-subnets to the relation, listing the
        # egress-subnets that have been granted access.
        # TODO: Meaningless in the k8s PostgreSQL charm. Can we have the
        # k8s Service limit connections? Do we want to?
        allowed_subnets = self.get_allowed_subnets(event.relation)
        allowed_units = self.get_allowed_units(event.relation)  # Legacy protocol, deprecated
        port = 5432

        # Publish connection details to the master.
        master = ConnectionString(
            host=master_ip,
            dbname=dbname,
            port=port,
            user=username,
            password=password,
            fallback_application_name=event.app.name,
        )
        standbys = ConnectionString(
            host=standbys_ip,
            dbname=dbname,
            port=port,
            user=username,
            password=password,
            fallback_application_name=event.app.name,
        )

        # Publish information on the unit relation data for backwards
        # compatibility, and to the application relation data for
        # modern clients.
        to_publish = [event.relation.data[self.unit]]
        if is_leader:
            to_publish.append(event.relation.data[self.app])
        for bucket in to_publish:
            bucket.update(
                {
                    # Echo back data to clients so they know their
                    # requested changes have been made.
                    "database": req.database,
                    "roles": req.sroles,
                    "extensions": req.sextensions,
                    "allowed-subnets": allowed_subnets,
                    # Connection strings, so that clients can connect.
                    "master": str(master),
                    "standbys": str(standbys),
                    # Charm only supports PG 12. If we support other versions,
                    # we would need to somehow extract this information from
                    # the image or a running pod. Pods could label themselves
                    # with the baked in version.
                    "version": str(PG_MAJOR),
                    # Legacy protocol for antique clients, deprecated.
                    "host": master_ip if is_leader else standbys_ip,
                    "state": "master" if is_leader else "hot standby",
                    "port": str(port),
                    "user": username,
                    "password": password,
                    "allowed-units": allowed_units,
                }
            )

    def db_password(self, username: str):
        if username not in self.passwords:
            if not self.unit.is_leader():
                return None
            self.passwords[username] = host.pwgen(40)
        return self.passwords[username]

    def get_allowed_subnets(self, relation: ops.model.Relation) -> str:
        subnets = set()
        for unit, reldata in relation.data.items():
            if isinstance(unit, ops.model.Unit) and not unit.name.startswith(self.model.app.name):
                # NB. egress-subnets is not always available.
                subnets.update(set(_csplit(reldata.get("egress-subnets", ""))))
        return ",".join(sorted(subnets))

    def get_allowed_units(self, relation: ops.model.Relation) -> str:
        return ",".join(
            sorted(
                unit.name
                for unit in relation.data
                if isinstance(unit, ops.model.Unit) and not unit.name.startswith(self.model.app.name)
            )
        )


class _ClientRequests:
    database: str = ""
    sroles: str = ""
    roles: list = None
    sextensions: str = ""
    extensions: list = None

    def __init__(self, event: ops.charm.RelationEvent) -> None:
        self.roles = []
        self.extensions = []
        # Inspect requests from the client. First look in Application
        # data for modern clients. Fall back to eventually consistent
        # unit data.
        buckets = [event.relation.data[event.app]]
        if event.unit:
            buckets.append(event.relation.data[event.unit])
        for bucket in buckets:
            database = bucket.get("database", "")
            sroles = bucket.get("roles", "")
            roles = list(_csplit(sroles))
            sextensions = bucket.get("extensions", "")
            extensions = list(_csplit(sextensions))
            if database or roles or extensions:
                self.database = database
                self.sroles = sroles
                self.roles = roles
                self.sextensions = sextensions
                self.extensions = extensions
                break


def _csplit(s) -> Iterable[str]:
    if s:
        for b in s.split(","):
            b = b.strip()
            if b:
                yield b


def _core_v1_api():
    cl = kubernetes.client.ApiClient()
    return kubernetes.client.CoreV1Api(cl)


def _fix_lp_1892255():
    # Remove os.environ.update when lp:1892255 is FIX_RELEASED.
    os.environ.update(
        dict(e.split("=") for e in Path("/proc/1/environ").read_text().split("\x00") if "KUBERNETES_SERVICE" in e)
    )
